/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef ASSEMBLYJOBPERFECT_CPP
#define ASSEMBLYJOBPERFECT_CPP


#include "AssemblyJobPerfect.h"

#include "AssemblyJob.h"
#include "AssemblyException.h"
#include "AlignmentFull.h"
#include <queue>
#include <string>
#include <iostream>
using namespace::std;

AssemblyJobPerfect::AssemblyJobPerfect(){}
AssemblyJobPerfect::AssemblyJobPerfect(set<ScoredSeq*>* seqs,
				       AssemblyJob::AssemblyStrandedness strandedness) :
  _wasRun(false),
  _inputRemoved(false),
  _novelRemoved(false),
  _strandedness(strandedness){
  _inputSeqs.insert(seqs->begin(),seqs->end());
}


AssemblyJobPerfect::~AssemblyJobPerfect(){
}



bool AssemblyJobPerfect::wasRun(){ return _wasRun; }

void AssemblyJobPerfect::allInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
}

void AssemblyJobPerfect::remainingInputSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobP::remainingInputSeqs; job hasn't run yet.");
  }
  if ( _inputRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobP::remainingInputSeqs; seqs were already removed.");
  }
  outsideSet->insert( _retainedSeqs.begin(), _retainedSeqs.end() );
}

void AssemblyJobPerfect::discardedInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _usedSeqs.begin(), _usedSeqs.end() );
}

void AssemblyJobPerfect::newlyAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobP::newlyAssembledSeqs; job hasn't run yet.");
  }
  if ( _novelRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobP::newlyAssembledSeqs; seqs were already removed.");
  }
  outsideSet->insert( _novelSeqs.begin(), _novelSeqs.end() );
}

void AssemblyJobPerfect::allAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobP::allAssembledSeqs; job hasn't run yet.");
  }
  if ( _inputRemoved or _novelRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobP::allAssembledSeqs; seqs were already removed.");
  }
  outsideSet->insert( _retainedSeqs.begin(), _retainedSeqs.end() );
  outsideSet->insert( _novelSeqs.begin(), _novelSeqs.end() );
}
/*
void AssemblyJobPerfect::allInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
  outsideSet->insert( _usedSeqs.begin(), _usedSeqs.end() );
}

void AssemblyJobPerfect::remainingInputSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobIterative::remainingInputSeqs; job hasn't run yet.");
  }
  if ( _inputRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobIterative::remainingInputSeqs; seqs were already removed.");
  }
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
}

void AssemblyJobPerfect::discardedInputSeqs( set<ScoredSeq*>* outsideSet ){
  outsideSet->insert( _usedSeqs.begin(), _usedSeqs.end() );
}
void AssemblyJobPerfect::newlyAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobIterative::newlyAssembledSeqs; job hasn't run yet.");
  }
  if ( _novelRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobIterative::newlyAssembledSeqs; seqs were already removed.");
  }
  outsideSet->insert( _novelSeqs.begin(), _novelSeqs.end() );
}

void AssemblyJobPerfect::allAssembledSeqs( set<ScoredSeq*>* outsideSet ){
  if (! _wasRun){
    throw AssemblyException::LogicError("in AssemblyJobIterative::allAssembledSeqs; job hasn't run yet.");
  }
  if ( _inputRemoved or _novelRemoved ){
    throw AssemblyException::LogicError("in AssemblyJobIterative::allAssembledSeqs; seqs were already removed.");
  }
  outsideSet->insert( _inputSeqs.begin(), _inputSeqs.end() );
  outsideSet->insert( _novelSeqs.begin(), _novelSeqs.end() );
}
*/




void AssemblyJobPerfect::OK(){
  if (! _wasRun ){
    if ( _novelSeqs.size() != 0 or _allAssembledSeqs.size() != 0 or _remainingSeqs.size() != 0 ){
      throw AssemblyException::LogicError("AssemblyJobPerfect::OK; there should be no product seqs if job hasn't run.");
    }
    if ( _novelRemoved or _inputRemoved ){
      throw AssemblyException::LogicError("AssemblyJobPerfect::OK; seqs can't have been removed if job hasn't run.");
    }
  }
}







void AssemblyJobPerfect::makeShallow(bool shallowDelete){
  if (_wasRun){
    throw AssemblyException::CallingError("AJNull::makeShallow can't be called after the job was run.");
  }
  set<ScoredSeq*> shallowCopies;
  // buffer all input, then copy
  for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
    (*it)->bottomBuffer();
  }
  for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
    shallowCopies.insert( (*it)->shallowCopy() );
    (*it)->deepUnbuffer();
  }
  if ( shallowDelete){
    for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
      delete (*it);
    }
  }
  // replace the input contents with the shallow copies
  _inputSeqs.clear();
  _inputSeqs.insert( shallowCopies.begin(), shallowCopies.end() );
}

void AssemblyJobPerfect::runJob(AssemblyJob::AssemblyThreadedness threadedness){
  if (! _wasRun){

    // get all of the lengths; only run the collapse for sets of similar-length sequences
    // STEP 1: organize input data by sequence length
    map<long, set<ScoredSeq*>* > lenToSeqSet;
    for (set<ScoredSeq*>::iterator inputIt = _inputSeqs.begin(); inputIt != _inputSeqs.end(); ++inputIt){
      ScoredSeq* inSeq = (*inputIt);
      map<long, set<ScoredSeq*>* >::iterator lenToSeqsIt = lenToSeqSet.find( inSeq->size() );
      if ( lenToSeqsIt == lenToSeqSet.end() ){
	set<ScoredSeq*>* newEntry = new set<ScoredSeq*>;
	newEntry->insert( inSeq );
	lenToSeqSet.insert( pair<int, set<ScoredSeq*>* > (inSeq->size(), newEntry) );
      } else {
	lenToSeqsIt->second->insert( inSeq );
      }
    }

    // organize into arrays to facilitate threading
    set<ScoredSeq*>** indexToSeqSet = new set<ScoredSeq*>*[ lenToSeqSet.size() + 1 ];
    long numLengthSets = 0;
    for (map<long, set<ScoredSeq*>* >::iterator lenToSeqsIt = lenToSeqSet.begin(); lenToSeqsIt != lenToSeqSet.end(); ++lenToSeqsIt){
      indexToSeqSet[numLengthSets] = lenToSeqsIt->second;
      numLengthSets++;
    }

    // do the collapsing
    if (threadedness == AssemblyJob::threaded){
      #pragma omp parallel for schedule(dynamic)
      for (long n = 0; n < numLengthSets; ++n){
	subAssembly(indexToSeqSet[n]);
      }
    } else {
      for (long n = 0; n < numLengthSets; ++n){
	subAssembly(indexToSeqSet[n]);
      }
    }

    // STEP 3: clean up
    // re-consolidate and sort the sequences
    for (long n = 0; n < numLengthSets; ++n){
      for (set<ScoredSeq*>::iterator it = indexToSeqSet[n]->begin(); it != indexToSeqSet[n]->end(); ++it){
	if ( _inputSeqs.count(*it) == 0 ){ _novelSeqs.insert(*it); }
	else { _retainedSeqs.insert(*it); }
      }
      delete indexToSeqSet[n];
    }
    delete [] indexToSeqSet;
    // figure out which of the input seqs was used
    for (set<ScoredSeq*>::iterator it = _inputSeqs.begin(); it != _inputSeqs.end(); ++it){
      if (_retainedSeqs.count(*it) == 0){ _usedSeqs.insert(*it); }
    }
    _wasRun = true;
  }
  //OK();
}



void AssemblyJobPerfect::subAssembly(set<ScoredSeq*>* monoLenSeqSet){
  // it only needs to modify the set if there are multiple entries, otherwise the seq will not be changed
  if (monoLenSeqSet->size() > 1){
 
    map<string,ScoredSeq*> stringToSeq; // the condensed set, indexed by seqeuence
    map<string,string> stringToRc; // for looking for minus-strand matches
    for (set<ScoredSeq*>::iterator inputIt = monoLenSeqSet->begin(); inputIt != monoLenSeqSet->end(); ++inputIt){
      bool senseMatch = false;
      bool antisMatch = false;

      // try for a sense match
      char* keySeq = (*inputIt)->getSeq('+');
      map<string,ScoredSeq*>::iterator matchIt = stringToSeq.find(keySeq);
      if ( matchIt!=stringToSeq.end() ){
	senseMatch = true;
	ScoredSeq* oldContig = matchIt->second;
	ScoredSeq* newContig = subAssemblyHelper( (*inputIt), oldContig , '+' );
	if (monoLenSeqSet->count(oldContig)==0){ delete oldContig; }
	matchIt->second = newContig;
      }

      // if appropriate, try for an antisense match
      if (_strandedness == AssemblyJob::doubleStranded and (! senseMatch) ){
	map<string,string>::iterator matchRcIt = stringToRc.find(keySeq);
	if ( matchRcIt != stringToRc.end() ){
	  antisMatch = true;
	  matchIt = stringToSeq.find( matchRcIt->second );
	  ScoredSeq* oldContig = matchIt->second;
	  ScoredSeq* newContig = subAssemblyHelper( (*inputIt), oldContig , '-' );
	  if (monoLenSeqSet->count(oldContig)==0){ delete oldContig; }
	  matchIt->second = newContig;
	}
      }

      // no match was found, so just add the seq
      if ( (! senseMatch) and (! antisMatch) ){
	stringToSeq.insert( pair<string,ScoredSeq*> (keySeq, (*inputIt) ) );
	if (_strandedness == AssemblyJob::doubleStranded){
	  char* revKeySeq = (*inputIt)->getSeq('-');
	  stringToRc.insert( pair<string,string>(revKeySeq,keySeq) );
	  delete [] revKeySeq;
	}
      }
      delete [] keySeq;
    }

    monoLenSeqSet->clear();
    for (map<string,ScoredSeq*>::iterator seqIt = stringToSeq.begin(); seqIt != stringToSeq.end(); ++seqIt){
      monoLenSeqSet->insert( seqIt->second );
    }
  }
}

// returned seq will be on the same strand as matchSeq
ScoredSeq* AssemblyJobPerfect::subAssemblyHelper(ScoredSeq* newInput, ScoredSeq* matchSeq, char sense){
  // create the new (combined) ScoredSeq
  Alignment* al = new AlignmentFull( matchSeq, newInput, sense ); // order is correct here to get the same sense as the matchSeq
  ScoredSeq* newContig = al->combine();
  delete al;
  return newContig;
}


#endif
