/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef READPAIRFILTER_H
#define READPAIRFILTER_H

#include <set>
#include "ScoredSeq.h"
#include "ParameterizedInitialFile.h"
#include "ScoredSeqCollectionBwt.h"
using namespace::std;

// this is a pure virtual class that will allow the contigs generated
// at the end of each cycle to be filtered according to some arbitrary
// criteria set by the implementing classes, which are below.

class ReadPairFilter {
 public:
  virtual ~ReadPairFilter();
  virtual ReadPairFilter* copy() = 0;
  virtual bool isReadOk(ScoredSeq* read) = 0;
  virtual bool isPairOk(ScoredSeq* readA, ScoredSeq* readB) = 0;
};


// i am including the implementing classes in this file

// this one always returns true
class ReadPairFilterNull : public ReadPairFilter {
 public:
  ReadPairFilterNull();
  ~ReadPairFilterNull();
  ReadPairFilterNull* copy();
  bool isReadOk(ScoredSeq* read);
  bool isPairOk(ScoredSeq* readA, ScoredSeq* readB);
};

// makes sure a seq has enough score counts
class ReadPairFilterHomoPolymer : public ReadPairFilter {
 public:
  ReadPairFilterHomoPolymer();
  ReadPairFilterHomoPolymer(long maxHomoPolymer);
  ~ReadPairFilterHomoPolymer();
  ReadPairFilterHomoPolymer* copy();
  bool isReadOk(ScoredSeq* read);
  bool isPairOk(ScoredSeq* readA, ScoredSeq* readB);
 private:
  long _maxHomoPolymer;
};

// gets rid of sequences that match some target
class ReadPairFilterAvoidSeqs : public ReadPairFilter {
 public:
  ReadPairFilterAvoidSeqs();
  ReadPairFilterAvoidSeqs(set<ScoredSeq*>* seqsToAvoid, float minFractId);
  ~ReadPairFilterAvoidSeqs();
  ReadPairFilterAvoidSeqs* copy();
  bool isReadOk(ScoredSeq* read);
  bool isPairOk(ScoredSeq* readA, ScoredSeq* readB);
 private:
  ScoredSeqCollectionBwt* _seqsToAvoid;
};

// imposes multiple tests
class ReadPairFilterMulti : public ReadPairFilter {
 public:
  ReadPairFilterMulti();
  ReadPairFilterMulti(ReadPairFilter** testArray, int numTests);
  ~ReadPairFilterMulti();
  ReadPairFilterMulti* copy();
  bool isReadOk(ScoredSeq* read);
  bool isPairOk(ScoredSeq* readA, ScoredSeq* readB);
 private:
  ReadPairFilter** _testArray;
  int _numTests;
};


#endif


