/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SCOREDSEQCOLLECTIONBWT_CPP
#define SCOREDSEQCOLLECTIONBWT_CPP

#include "ScoredSeqCollectionBwt.h"

#include "AssemblyException.h"
#include "AlignmentNull.h"
#include "AlignmentUngapped.h"
#include "ScoredSeqFlip.h"
#include <iostream>
#include <omp.h>
using namespace::std;


ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(){}

ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, float fractId) : 
  _asi(0),
  _gapMode(ungapped),
  _fractId(fractId),
  _minOverlap(0),
  _penalizeEdgeGaps(true),
  _usingMaxOverlap(false),
  _maxOverlap(-1) {
  constructorHelper(inputSeqs);
}
ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, float fractId, long minOverlap) : 
  _asi(0),
  _gapMode(ungapped),
  _fractId(fractId),
  _minOverlap(minOverlap),
  _penalizeEdgeGaps(false),
  _usingMaxOverlap(false),
  _maxOverlap(-1) {
  constructorHelper(inputSeqs);
}
ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, float fractId, long minOverlap, long maxOverlap) : 
  _asi(0),
  _gapMode(ungapped),
  _fractId(fractId),
  _minOverlap(minOverlap),
  _penalizeEdgeGaps(false),
  _usingMaxOverlap(true),
  _maxOverlap(maxOverlap) {
  constructorHelper(inputSeqs);
}
ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, DynamicProgrammingAligner * asi) :
  _gapMode(gapped),
  _fractId(asi->getFractId()),
  _minOverlap(asi->getMinOverlap()),
  _penalizeEdgeGaps(false),
  _usingMaxOverlap(false),
  _maxOverlap(-1) {
  _asi = new DynamicProgrammingAligner(asi);
  constructorHelper(inputSeqs);
}
ScoredSeqCollectionBwt::ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, DynamicProgrammingAligner * asi, long maxOverlap) :
  _gapMode(gapped),
  _fractId(asi->getFractId()),
  _minOverlap(asi->getMinOverlap()),
  _penalizeEdgeGaps(false),
  _usingMaxOverlap(true),
  _maxOverlap(maxOverlap) {
  _asi = new DynamicProgrammingAligner(asi);
  constructorHelper(inputSeqs);
}

void ScoredSeqCollectionBwt::constructorHelper(set<ScoredSeq*>* inputSeqs){
  // by default, edge scaling will happen.  it can be disabled to save
  // time during read mapping.
  _edgeScalingEnabled = true;

  // minOverlap cannot be zero for this class; sub-string words must be defined
  // and statistically scaled for this overlap minimum, so a hard bottom line
  // needs to be defined here
  long minOverlapLimit = 4;
  if (_minOverlap < minOverlapLimit){ _minOverlap = minOverlapLimit; }

  // fractId must be at least that expected for random sequence matches
  if (_fractId < 0.25){ _fractId = 0.25; }

  // some data that will be useful to have pre-computed
  _maxFractMis = 1.0 - _fractId;
  long matchScore = 1;
  long misScore = -1;
  _scoreAsm = new AlignmentScoreMatrix(matchScore,misScore,misScore,misScore);
  _matchCountAsm = new AlignmentScoreMatrix(1,0,0,0);
  _misCountAsm = new AlignmentScoreMatrix(0,1,1,1); // gaps count as mismatches (terminal gaps will exist)
  _allSeqs.insert( inputSeqs->begin(), inputSeqs->end() );

  // figure out the windows and make the bins
  long seqCount = 0;
  long seqLenSum = 0;
  for (set<ScoredSeq*>::iterator it = inputSeqs->begin(); it != inputSeqs->end(); ++it){
    seqCount++;
    seqLenSum += (*it)->size() + 1; // +1 because of the separator/end char
  }
  if (seqCount==0){
    _binSize = 0; // this value is only possible in this case
    _maxBin = 0;
  } else {
    _binSize = seqLenSum / seqCount + 1; // +1 to round up
    _maxBin = seqLenSum / _binSize;
  }
  _binToSeqs = new set<ScoredSeqLocus*>*[_maxBin+1];

  
  for (long b = 0; b <= _maxBin; ++b){
    _binToSeqs[b] = new set<ScoredSeqLocus*>;
  }

  // convert the full text to a numeric array, filling in the bins
  if (seqLenSum==0){ seqLenSum = 1; } // at least an end char will be added
  _textSize = seqLenSum;

  // a local array of text length
  long* textAsNums = new long[seqLenSum];

  long currentIndex = 0;
  _numContigLoci = inputSeqs->size();
  if (_numContigLoci > 0){
    _orderedContigLoci = new ScoredSeqLocus*[ _numContigLoci ];
  }

  long currentOffset = 0;
  for (set<ScoredSeq*>::iterator it = _allSeqs.begin(); it != _allSeqs.end(); ++it){
    // add the sequence, preceeded by a separator if necessary
    if ( it != _allSeqs.begin() ){
      char* separator = new char[2];
      separator[0] = '&';
      separator[1] = '\0';
      burrowsWheelerUtilities::makeNumString( separator, long(1), currentOffset, textAsNums );
      delete [] separator;
      currentOffset++;
    }
    ScoredSeq* seq = (*it);
    char* seqString = seq->getSeq('+');
    burrowsWheelerUtilities::makeNumString( seqString, seq->size(), currentOffset, textAsNums );
    delete [] seqString;

    // fill the overlapping bins with this ScoredSeq in Locus form
    long firstBin = currentOffset / _binSize;
    long lastBinP1 = ( currentOffset + seq->size() ) / _binSize + 1;
    ScoredSeqLocus* seqAsLocus = new ScoredSeqLocus(seq, currentOffset, currentOffset + seq->size() - 1, currentIndex);
    for (long bin = firstBin; bin < lastBinP1; ++bin){ _binToSeqs[bin]->insert( seqAsLocus ); }

    _orderedContigLoci[currentIndex] = seqAsLocus;
    currentIndex++;

    currentOffset += seq->size();
  }

  // end the sequence with an end char (even if there were no sequence entries)
  char* endChar = new char[2];
  endChar[0] = '#';
  endChar[1] = '\0';
  burrowsWheelerUtilities::makeNumString( endChar, long(1), currentOffset, textAsNums );
  delete [] endChar;

  // make the burrows-wheeler transform
  _sortedSuffixes = new long[seqLenSum];
  long alphaSize = burrowsWheelerUtilities::getAlphabetSize();
  burrowsWheelerUtilities::sortSuffixes( textAsNums, _sortedSuffixes, seqLenSum, alphaSize );
  _bwTransform = new long[seqLenSum];
  burrowsWheelerUtilities::bwTransform(textAsNums, _sortedSuffixes, seqLenSum, _bwTransform);

  delete [] textAsNums;

  // make the accessory data structures for alignment
  _occCounts = new long[seqLenSum];
  _tableC = burrowsWheelerUtilities::fillBwtHelperArrays(_bwTransform, seqLenSum, alphaSize, _occCounts);
  //OK();
}



void ScoredSeqCollectionBwt::OK(){
  if ( _fractId > 1 ){
    throw AssemblyException::ArgError("fraction ID is a fraction; cannot be greater than 1");
  }
  if (_allSeqs.size()==0){
    if (_textSize != 1){ throw AssemblyException::LogicError("BWT text len is 1 if there are no seqs"); }
  } else {
    long seqLenSum = 0;
    for (set<ScoredSeq*>::iterator it = _allSeqs.begin(); it != _allSeqs.end(); ++it){
      seqLenSum += (*it)->size() + 1; // +1 because of the separator/end char
    }
    if (_textSize != seqLenSum){ throw AssemblyException::LogicError("BWT text len is wrong"); }
  }
}



ScoredSeqCollectionBwt::~ScoredSeqCollectionBwt(){
  delete _asi;
  delete _matchCountAsm;
  delete _misCountAsm;
  delete _scoreAsm;
  delete [] _sortedSuffixes;
  delete [] _bwTransform;
  delete [] _occCounts;
  delete [] _tableC;
  for (long n = 0; n < _numContigLoci; ++n){ delete _orderedContigLoci[n]; }
  if (_numContigLoci > 0){ delete [] _orderedContigLoci; }

  for (long b = 0; b <= _maxBin; ++b){ delete _binToSeqs[b]; }
  delete [] _binToSeqs;
}




ScoredSeqCollectionBwt * ScoredSeqCollectionBwt::copy(){
  // uses a private constructor to efficiently copy stuff
  throw AssemblyException::ImplementationError("not implemented");
}

float ScoredSeqCollectionBwt::getFractId(){ return _fractId; }



// DEFAULT: edge scaling is enabled
void ScoredSeqCollectionBwt::enableEdgeScaling(){ _edgeScalingEnabled = true; }
void ScoredSeqCollectionBwt::disableEdgeScaling(){ _edgeScalingEnabled = false; }



bool ScoredSeqCollectionBwt::contains(ScoredSeq* s){
  //OK();
  return ( _allSeqs.find(s) != _allSeqs.end() );
}



void ScoredSeqCollectionBwt::getSeqs( set<ScoredSeq*>* allSeqs ){
  //OK();
  allSeqs->insert( _allSeqs.begin(), _allSeqs.end() );
}





void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* matchTest){
  set<Alignment*> tempMatches;
  getMatchesPrivate(&tempMatches, seq, matchTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}

void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* matchTest){
  set<Alignment*> tempMatches;
  getMatchesPrivate(&tempMatches, seq, sense, matchTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}


void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq){
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getMatchesPrivate(&tempMatches, seq, seqTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}

void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense){
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getMatchesPrivate(&tempMatches, seq, sense, seqTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}


// min overlap is specified
void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* matchTest,
					long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  getMatchesPrivate(&tempMatches, seq, matchTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}
void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* matchTest,
					long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  getMatchesPrivate(&tempMatches, seq, sense, matchTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}
void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq,
					long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getMatchesPrivate(&tempMatches, seq, seqTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}
void ScoredSeqCollectionBwt::getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense,
					long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getMatchesPrivate(&tempMatches, seq, sense, seqTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}


// ASSUMES: matches is empty; otherwise, it will be modified
// WILL RETURN MULTIPLE DISTINCT ALIGNMENTS TO A SINGLE TARGET
void ScoredSeqCollectionBwt::getMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* matchTest, long minOverlap){
  //OK();
  getMatchesPrivate(matches, seq, '+', matchTest, minOverlap);
  getMatchesPrivate(matches, seq, '-', matchTest, minOverlap);
  //OK();
}


// REQUIRES: all members of "matches" (if any) have the same score
void ScoredSeqCollectionBwt::getMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* matchTest, long minOverlap){
  bufferSeqs();

  long querySize = seq->size();
  AlignmentMaker * alMaker;
  if (_penalizeEdgeGaps){
    long maxMismatch = long(float(querySize) * _maxFractMis); // keep constant; this isn't best matches
    alMaker = new AlMakerPenalizeEdgeA(_scoreAsm, -10, _maxFractMis); // value in alMaker
  } else {
    alMaker = new AlMakerNoEdge(_scoreAsm, -10, _maxFractMis); // value in alMaker
  }

  // get the numeric representation of the query
  ScoredSeq* seqFlip = ScoredSeqFlip::getFlip(seq, sense);

  // for each matched contig, keeps track of the num hits with any
  // given offset
  OffsetTracker** contigToOffsets = getOffsets(seq,sense,minOverlap,matchTest);

  // make alignments
  long hitIndex = 0;
  while (contigToOffsets[hitIndex] != NULL){
    OffsetTracker* tracker = contigToOffsets[hitIndex];
    hitIndex++;
    ScoredSeq* contig = tracker->targetContig();
    long* allOffsets = tracker->getOffsets();
    long numOffsets = tracker->numOffsets();
    delete tracker;
    set<long> adequateOffsets;
    for (long offIndex = 0; offIndex < numOffsets; ++offIndex){
      // determine how long the alignment will be??  i will need to do that to make sure
      // that only legit alignments are kept.
      long offset = allOffsets[offIndex];
      long seqStart;
      if (offset < 0){ seqStart = 0; }
      else { seqStart = offset; }
      long seqEnd;
      if (offset + contig->size() >= seq->size()){ seqEnd = seq->size(); }
      else { seqEnd = offset + contig->size(); }
      long alLength = seqEnd - seqStart;
      if (alLength < 0){ throw AssemblyException::LogicError("SSCBwt::getMatches, alLength should be > 0"); }
      if ( (_penalizeEdgeGaps or alLength >= minOverlap) and ( (! _usingMaxOverlap) or alLength <= _maxOverlap) ){
	adequateOffsets.insert(offset);
      }
    }
    delete [] allOffsets;

    // this will be different for gapped versus ungapped alignments
    if (_gapMode == ungapped){ // UNGAPPED ALIGNMENT
      for (set<long>::iterator offsetIt = adequateOffsets.begin(); offsetIt != adequateOffsets.end(); ++offsetIt){
	long offset = *offsetIt;

	Alignment* alignCand = alMaker->makeAlignment(seqFlip, contig, '+', offset);

	if ( alignCand->isNull() ){ delete alignCand; }
	else {
	  if (sense == '+'){
	    alignCand->seqReplace(seq, alignCand->seqB());
	    matches->insert(alignCand);
	  } else {
	    matches->insert( alignCand->copyRcSeqA(seq) );
	    //matches->insert( reverseAlignmentSeqA(alignCand, seq) );
	    delete alignCand;
	  }
	}
      }
    } else { // GAPPED ALIGNMENT
      if (adequateOffsets.size() > 0){
	Alignment* alignCand = _asi->align(seqFlip,contig,'+',&adequateOffsets);
	if ( alignCand->isNull() ){ delete alignCand; }
	else {
	  if (sense == '+'){
	    alignCand->seqReplace(seq, alignCand->seqB());
	    matches->insert(alignCand);
	  } else {
	    matches->insert( alignCand->copyRcSeqA(seq) );
	    //matches->insert( reverseAlignmentSeqA(alignCand, seq) );
	    delete alignCand;
	  }
	}
      }
    }
  }
  delete [] contigToOffsets;

  delete alMaker;
  delete seqFlip;
  //if (currentAsi != asi){ delete currentAsi; }
}



void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* seqTest){
  set<Alignment*> tempMatches;
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, seqTest, alignTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete alignTest;
}

void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* seqTest){
  set<Alignment*> tempMatches;
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, alignTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete alignTest;
}


void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq){
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, seqTest, alignTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
  delete alignTest;
}

void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense){
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, alignTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
  delete alignTest;
}




void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* seqTest, MatchAlignTest* alignTest){
  set<Alignment*> tempMatches;
  getBestMatchesPrivate(&tempMatches, seq, seqTest, alignTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}

void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* seqTest, MatchAlignTest* alignTest){
  set<Alignment*> tempMatches;
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, alignTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchAlignTest* alignTest){
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, seqTest, alignTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchAlignTest* alignTest){
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, alignTest, _minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}



// min overlap is specified
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* seqTest,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, seqTest, alignTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete alignTest;
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* seqTest,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, alignTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete alignTest;
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, seqTest, alignTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
  delete alignTest;
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, alignTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
  delete alignTest;
}


// align tests
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* seqTest, MatchAlignTest* alignTest,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  getBestMatchesPrivate(&tempMatches, seq, seqTest, alignTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense,
					    MatchSeqTest* seqTest, MatchAlignTest* alignTest,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<Alignment*> tempMatches;
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, alignTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchAlignTest* alignTest,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, seqTest, alignTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}
void ScoredSeqCollectionBwt::getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchAlignTest* alignTest,
					    long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  set<ScoredSeq*> okTargets; // a dummy container just to fulfill the specs
  set<Alignment*> tempMatches;
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  getBestMatchesPrivate(&tempMatches, seq, sense, seqTest, alignTest, minOverlap);
  matches->insert( tempMatches.begin(), tempMatches.end() );
  delete seqTest;
}


// THE BEGINNING OF HASMATCH

bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, MatchSeqTest* seqTest){
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, seqTest, alignTest, _minOverlap);
  delete alignTest;
  return hasMatch;
}

bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, MatchSeqTest* seqTest){
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, sense, seqTest, alignTest, _minOverlap);
  delete alignTest;
  return hasMatch;
}


bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq){
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, seqTest, alignTest, _minOverlap);
  delete seqTest;
  delete alignTest;
  return hasMatch;
}

bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense){
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, sense, seqTest, alignTest, _minOverlap);
  delete seqTest;
  delete alignTest;
  return hasMatch;
}




bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, MatchSeqTest* seqTest, MatchAlignTest* alignTest){
  return hasMatchPrivate(seq, seqTest, alignTest, _minOverlap);
}

bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, MatchSeqTest* seqTest, MatchAlignTest* alignTest){
  return hasMatchPrivate(seq, sense, seqTest, alignTest, _minOverlap);
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, MatchAlignTest* alignTest){
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, seqTest, alignTest, _minOverlap);
  delete seqTest;
  return hasMatch;
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, MatchAlignTest* alignTest){
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, sense, seqTest, alignTest, _minOverlap);
  delete seqTest;
  return hasMatch;
}



// min overlap is specified
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, MatchSeqTest* seqTest, long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, seqTest, alignTest, minOverlap);
  delete alignTest;
  return hasMatch;
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, MatchSeqTest* seqTest, long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, sense, seqTest, alignTest, minOverlap);
  delete alignTest;
  return hasMatch;
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, seqTest, alignTest, minOverlap);
  delete seqTest;
  delete alignTest;
  return hasMatch;
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  MatchAlignTest* alignTest = MatchAlignTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, sense, seqTest, alignTest, minOverlap);
  delete seqTest;
  delete alignTest;
  return hasMatch;
}


// align tests
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, MatchSeqTest* seqTest, MatchAlignTest* alignTest,
				      long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  return hasMatchPrivate(seq, seqTest, alignTest, minOverlap);
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, MatchSeqTest* seqTest, MatchAlignTest* alignTest,
				      long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  return hasMatchPrivate(seq, sense, seqTest, alignTest, minOverlap);
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, MatchAlignTest* alignTest, long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, seqTest, alignTest, minOverlap);
  delete seqTest;
  return hasMatch;
}
bool ScoredSeqCollectionBwt::hasMatch(ScoredSeq* seq, char sense, MatchAlignTest* alignTest, long minOverlap, MinOvlStringency ovlStringency){
  if (ovlStringency == softMinOvl){ minOverlap = long(float(minOverlap) * _fractId); }
  MatchSeqTest* seqTest = MatchSeqTest::getNullTest();
  bool hasMatch = hasMatchPrivate(seq, sense, seqTest, alignTest, minOverlap);
  delete seqTest;
  return hasMatch;
}


// THE END OF HASMATCH



// ASSUMES: matches is empty; otherwise, it will be modified
// WILL RETURN MULTIPLE DISTINCT ALIGNMENTS TO A SINGLE TARGET
void ScoredSeqCollectionBwt::getBestMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq,
						   MatchSeqTest* seqTest, MatchAlignTest* alignTest,
						   long minOverlap){
  //OK();
  // for each matched contig, keeps track of the num hits with any
  // given offset
  OffsetTracker** ctoPlus = getOffsets(seq,'+',minOverlap,seqTest);
  OffsetTracker** ctoMinus = getOffsets(seq,'-',minOverlap,seqTest);

  getBestMatchesPrivateUngapped(ctoPlus, matches, seq, '+', seqTest, alignTest, minOverlap);
  getBestMatchesPrivateUngapped(ctoMinus, matches, seq, '-', seqTest, alignTest, minOverlap);
  if (_gapMode == gapped){
    long bestScoreSoFar = 0;
    if (matches->size() > 0){ bestScoreSoFar = (*(matches->begin()))->score(_scoreAsm,_penalizeEdgeGaps); }
    for (set<Alignment*>::iterator it = matches->begin(); it != matches->end(); ++it){ delete *it; }
    matches->clear();
    getBestMatchesPrivateGapped(ctoPlus, matches, seq, '+', seqTest, alignTest, minOverlap, bestScoreSoFar);
    getBestMatchesPrivateGapped(ctoMinus, matches, seq, '-', seqTest, alignTest, minOverlap, bestScoreSoFar);
  }

  // delete the offset tracker arrays and their contents
  long hitIndex = 0;
  while (ctoPlus[hitIndex] != NULL){
    delete ctoPlus[hitIndex];
    hitIndex++;
  }
  delete [] ctoPlus;
  hitIndex = 0;
  while (ctoMinus[hitIndex] != NULL){
    delete ctoMinus[hitIndex];
    hitIndex++;
  }
  delete [] ctoMinus;
  //OK();
}
// REQUIRES: all members of "matches" (if any) have the same score
void ScoredSeqCollectionBwt::getBestMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq, char sense,
						   MatchSeqTest* seqTest, MatchAlignTest* alignTest,
						   long minOverlap){
  // for each matched contig, keeps track of the num hits with any
  // given offset
  OffsetTracker** contigToOffsets = getOffsets(seq,sense,minOverlap,seqTest);
  getBestMatchesPrivateUngapped(contigToOffsets, matches, seq, sense, seqTest, alignTest, minOverlap);
  if (_gapMode == gapped){
    long bestScoreSoFar = 0;
    if (matches->size() > 0){ bestScoreSoFar = (*(matches->begin()))->score(_scoreAsm,_penalizeEdgeGaps); }
    for (set<Alignment*>::iterator it = matches->begin(); it != matches->end(); ++it){ delete *it; }
    matches->clear();
    getBestMatchesPrivateGapped(contigToOffsets, matches, seq, sense, seqTest, alignTest, minOverlap, bestScoreSoFar);
  }
  // delete array and contents
  long hitIndex = 0;
  while (contigToOffsets[hitIndex] != NULL){
    delete contigToOffsets[hitIndex];
    hitIndex++;
  }
  delete [] contigToOffsets;
  //OK();
}




// ASSUMES: matches is empty; otherwise, it will be modified
// WILL RETURN MULTIPLE DISTINCT ALIGNMENTS TO A SINGLE TARGET
bool ScoredSeqCollectionBwt::hasMatchPrivate(ScoredSeq* seq,
					     MatchSeqTest* seqTest, MatchAlignTest* alignTest,
					     long minOverlap){
  set<Alignment*> matches;
  // for each matched contig, keeps track of the num hits with any
  // given offset
  OffsetTracker** ctoPlus = getOffsets(seq,'+',minOverlap,seqTest);
  OffsetTracker** ctoMinus = getOffsets(seq,'-',minOverlap,seqTest);

  getBestMatchesPrivateUngapped(ctoPlus, &matches, seq, '+', seqTest, alignTest, minOverlap, true);
  if (matches.size()==0){ getBestMatchesPrivateUngapped(ctoMinus, &matches, seq, '-', seqTest, alignTest, minOverlap, true); }
  if (matches.size()==0 and _gapMode == gapped){
    long bestScoreSoFar = 0;
    getBestMatchesPrivateGapped(ctoPlus, &matches, seq, '+', seqTest, alignTest, minOverlap, bestScoreSoFar, true);
    if (matches.size()==0){ getBestMatchesPrivateGapped(ctoMinus, &matches, seq, '-', seqTest, alignTest, minOverlap, bestScoreSoFar, true); }
  }

  // delete the offset tracker arrays and their contents
  long hitIndex = 0;
  while (ctoPlus[hitIndex] != NULL){
    delete ctoPlus[hitIndex];
    hitIndex++;
  }
  delete [] ctoPlus;
  hitIndex = 0;
  while (ctoMinus[hitIndex] != NULL){
    delete ctoMinus[hitIndex];
    hitIndex++;
  }
  delete [] ctoMinus;
  bool hasMatch = matches.size() > 0;
  for (set<Alignment*>::iterator it = matches.begin(); it != matches.end(); ++it){ delete *it; }
  return hasMatch;
}
// REQUIRES: all members of "matches" (if any) have the same score
bool ScoredSeqCollectionBwt::hasMatchPrivate(ScoredSeq* seq, char sense,
					     MatchSeqTest* seqTest, MatchAlignTest* alignTest,
					     long minOverlap){
  set<Alignment*> matches;
  // for each matched contig, keeps track of the num hits with any
  // given offset
  OffsetTracker** contigToOffsets = getOffsets(seq,sense,minOverlap,seqTest);
  getBestMatchesPrivateUngapped(contigToOffsets, &matches, seq, sense, seqTest, alignTest, minOverlap, true);
  if (matches.size()==0 and _gapMode == gapped){
    long bestScoreSoFar = 0;
    getBestMatchesPrivateGapped(contigToOffsets, &matches, seq, sense, seqTest, alignTest, minOverlap, bestScoreSoFar, true);
  }
  // delete array and contents
  long hitIndex = 0;
  while (contigToOffsets[hitIndex] != NULL){
    delete contigToOffsets[hitIndex];
    hitIndex++;
  }
  delete [] contigToOffsets;
  bool hasMatch = matches.size() > 0;
  for (set<Alignment*>::iterator it = matches.begin(); it != matches.end(); ++it){ delete *it; }
  return hasMatch;
}



// THESE WILL REPLACE THE SINGLE METHOD BELOW
// REQUIRES: all members of "matches" (if any) have the same score
void ScoredSeqCollectionBwt::getBestMatchesPrivateUngapped(OffsetTracker** contigToOffsets,
							   set<Alignment*>* matches, ScoredSeq* seq, char sense,
							   MatchSeqTest* seqTest, MatchAlignTest* alignTest, long minOverlap, bool justFirstMatch){
  bufferSeqs();

  if (matches->size() > 0 and justFirstMatch){
    throw AssemblyException::CallingError("SSCBwt:getBestMatchesPrivateUngapped should not be called for justOneMatch if one was already found");
  }

  // must compete with prior matches (if any), but zero is the baseline
  long bestScoreSoFar = 0;
  if (matches->size() > 0){ bestScoreSoFar = (*(matches->begin()))->score(_scoreAsm,_penalizeEdgeGaps); }

  long querySize = seq->size();
  AlignmentMaker * alMaker;
  if (_penalizeEdgeGaps){
    long maxMismatch = long(float(querySize) * _maxFractMis); // keep constant; this isn't best matches
    alMaker = new AlMakerPenalizeEdgeA(_scoreAsm, -10, _maxFractMis); // value in alMaker
  } else {
    alMaker = new AlMakerNoEdge(_scoreAsm, -10, _maxFractMis); // value in alMaker
  }

  // get the numeric representation of the query
  ScoredSeq* seqFlip = ScoredSeqFlip::getFlip(seq, sense);

  // make alignments
  long hitIndex = 0;
  while (contigToOffsets[hitIndex] != NULL and ( (! justFirstMatch) or matches->size() == 0) ){
    OffsetTracker* contigIt = contigToOffsets[hitIndex];
    hitIndex++;
    // get the subset of offsets that are good enough for alignment according
    // to the global parameters
    ScoredSeq* contig = contigIt->targetContig();

    long* allOffsets = contigIt->getOffsets();
    long numOffsets = contigIt->numOffsets();
    long* adequateOffsets = new long[ numOffsets + 1 ];
    long numAdequateOffsets = 0;
    for (long allIndex = 0; allIndex < numOffsets; ++allIndex){
      // determine how long the alignment will be??  i will need to do that to make sure
      // that only legit alignments are kept.
      long offset = allOffsets[allIndex];
      long seqStart;
      if (offset < 0){ seqStart = 0; }
      else { seqStart = offset; }
      long seqEnd;
      if (offset + contig->size() >= seq->size()){ seqEnd = seq->size(); }
      else { seqEnd = offset + contig->size(); }
      long alLength = seqEnd - seqStart;
      if (alLength < 0){ throw AssemblyException::LogicError("SSCBwt::getBestMatches, alLength should be > 0"); }
      if ( (_penalizeEdgeGaps or alLength >= minOverlap) and ( (! _usingMaxOverlap) or alLength <= _maxOverlap) ){
	adequateOffsets[numAdequateOffsets] = offset;
	numAdequateOffsets++;
      }
    }
    delete [] allOffsets;

    for (long adequateIndex = 0; adequateIndex < numAdequateOffsets; ++adequateIndex){
      // this will speed things up in the just-one-match case by preventing alignments from being generated
      if ( (! justFirstMatch) or matches->size() == 0){
	long offset = adequateOffsets[adequateIndex];

	Alignment* alignCand = alMaker->makeAlignment(seqFlip, contig, '+', offset);
	long newScore;
	if (! alignCand->isNull()){ newScore = alMaker->getScore(alignCand); }
        if ( alignCand->isNull() or (matches->size() > 0 and newScore < bestScoreSoFar)
	     or (! alignTest->alignIsOk(alignCand) ) ){
	  delete alignCand;
	} else {
          // replace the old alignments of the score is bigger; also, set the best score in
          // that case or if this is the first legit alignment
	  if (matches->size()==0 or newScore > bestScoreSoFar){
	    bestScoreSoFar = newScore;
	    for (set<Alignment*>::iterator oldIt = matches->begin(); oldIt != matches->end(); ++oldIt){ delete (*oldIt); }
	    matches->clear();
	    alMaker->setMinScore( newScore );
	  }
          // add the alignment to the set of matches
	  if (matches->size()==0 or newScore >= bestScoreSoFar){
	    if (sense == '+'){
	      alignCand->seqReplace(seq, alignCand->seqB());
	      matches->insert(alignCand);
	    } else {
	      matches->insert( alignCand->copyRcSeqA(seq) );
	      delete alignCand;
	    }
	  } else { delete alignCand; }
	}
      }
    }
    delete [] adequateOffsets;
  }

  delete alMaker;
  delete seqFlip;
}

// REQUIRES: all members of "matches" (if any) have the same score
void ScoredSeqCollectionBwt::getBestMatchesPrivateGapped(OffsetTracker** contigToOffsets,
							 set<Alignment*>* matches, ScoredSeq* seq, char sense,
							 MatchSeqTest* seqTest, MatchAlignTest* alignTest,
							 long minOverlap, long bestScoreSoFar, bool justFirstMatch){
  bufferSeqs();

  if (matches->size() > 0 and justFirstMatch){
    throw AssemblyException::CallingError("SSCBwt:getBestMatchesPrivateGapped should not be called for justOneMatch if one was already found");
  }

  // must compete with prior matches (if any), but zero is the baseline
  if (matches->size() > 0){ bestScoreSoFar = (*(matches->begin()))->score(_scoreAsm,_penalizeEdgeGaps); }

  long querySize = seq->size();
  AlignmentMaker * alMaker;
  if (_penalizeEdgeGaps){
    long maxMismatch = long(float(querySize) * _maxFractMis); // keep constant; this isn't best matches
    alMaker = new AlMakerPenalizeEdgeA(_scoreAsm, -10, _maxFractMis); // value in alMaker
  } else {
    alMaker = new AlMakerNoEdge(_scoreAsm, -10, _maxFractMis); // value in alMaker
  }

  // get the numeric representation of the query
  ScoredSeq* seqFlip = ScoredSeqFlip::getFlip(seq, sense);

  // make alignments
  long hitIndex = 0;
  while (contigToOffsets[hitIndex] != NULL and ( (! justFirstMatch) or matches->size() == 0) ){
    OffsetTracker* contigIt = contigToOffsets[hitIndex];
    hitIndex++;
    // get the subset of offsets that are good enough for alignment according
    // to the global parameters
    ScoredSeq* contig = contigIt->targetContig();

    long* allOffsets = contigIt->getOffsets();
    long numOffsets = contigIt->numOffsets();
    long* adequateOffsets = new long[ numOffsets + 1 ];
    long numAdequateOffsets = 0;
    for (long allIndex = 0; allIndex < numOffsets; ++allIndex){
      // determine how long the alignment will be??  i will need to do that to make sure
      // that only legit alignments are kept.
      long offset = allOffsets[allIndex];
      long seqStart;
      if (offset < 0){ seqStart = 0; }
      else { seqStart = offset; }
      long seqEnd;
      if (offset + contig->size() >= seq->size()){ seqEnd = seq->size(); }
      else { seqEnd = offset + contig->size(); }
      long alLength = seqEnd - seqStart;
      if (alLength < 0){ throw AssemblyException::LogicError("SSCBwt::getBestMatches, alLength should be > 0"); }
      if ( (_penalizeEdgeGaps or alLength >= minOverlap) and ( (! _usingMaxOverlap) or alLength <= _maxOverlap) ){
	adequateOffsets[numAdequateOffsets] = offset;
	numAdequateOffsets++;
      }
    }
    delete [] allOffsets;

    // set up an aligner that will dynamically set new minScores
    DynamicProgrammingAligner * localAsi = _asi->minScoreCopy(bestScoreSoFar);

    if (numAdequateOffsets > 0){
      Alignment* alignCand = localAsi->align(seqFlip,contig,'+',adequateOffsets,numAdequateOffsets);
      if ( alignCand->isNull() ){ delete alignCand; }
      else {
	long newBestScore = alignCand->score(_scoreAsm,_penalizeEdgeGaps);
	if (matches->size()==0 or newBestScore > bestScoreSoFar){
	  bestScoreSoFar = newBestScore;
	  DynamicProgrammingAligner * newLocalAsi = localAsi->minScoreCopy(bestScoreSoFar);
	  delete localAsi;
	  localAsi = newLocalAsi;
	  for (set<Alignment*>::iterator oldIt = matches->begin(); oldIt != matches->end(); ++oldIt){ delete (*oldIt); }
	  matches->clear();
	}
	if (matches->size()==0 or newBestScore >= bestScoreSoFar){
	  if (sense == '+'){
	    alignCand->seqReplace(seq, alignCand->seqB());
	    matches->insert(alignCand);
	  } else {
	    matches->insert( alignCand->copyRcSeqA(seq) );
	    //matches->insert( reverseAlignmentSeqA(alignCand, seq) );
	    delete alignCand;
	  }
	} else { delete alignCand; }
      }
    }
    delete localAsi;
    delete [] adequateOffsets;
  }

  delete alMaker;
  delete seqFlip;
}





ScoredSeqCollectionBwt::OffsetTracker** ScoredSeqCollectionBwt::getOffsets(ScoredSeq* seq, char sense, long minOverlap, MatchSeqTest* matchTest){

  // this is what I am collecting now so that I can sort the return values
  map<ScoredSeq*,long> hitToIndex;
  long currentIndex = 0;
  OffsetTracker** hitArray = new OffsetTracker*[ _allSeqs.size() ];

  long querySize = seq->size();
  long* queryNumArray = new long[querySize];
  char* seqString = seq->getSeq(sense);
  burrowsWheelerUtilities::makeNumString( seqString, seq->size(), 0, queryNumArray );
  delete [] seqString;

  // get the size and position of the sub-strings for which
  // matches will be sought
  set<WordQuery*> wordQuerySet;
  getSubtextWordQueries(seq, &wordQuerySet, sense, minOverlap);

  for (set<WordQuery*>::iterator it = wordQuerySet.begin(); it != wordQuerySet.end(); ++it){
    WordQuery* wordQuery = *it;
    long stSize = wordQuery->_size;
    long* subtext = new long[stSize];

    // build the subtext num array
    long startNum = wordQuery->_start;
    for (int n = 0; n < stSize; ++n){ subtext[n] = queryNumArray[n+startNum]; }

    // do the search with the sub-query word
    // deal with possible redundant matchLoci if multiple words match
    long* matchArray = burrowsWheelerUtilities::bwtCountAndFind(subtext, stSize,
								_bwTransform, _sortedSuffixes, _textSize,
								_tableC, _occCounts);

    // keep track of the starts AND how many non-overlapping windows have matched
    long numMatches = matchArray[0];
    for (long n = 1; n <= numMatches; ++n){
      long hitStart = matchArray[n];
      ScoredSeqLocus* contigLocus = _orderedContigLoci[ getOverlapContigIndex(hitStart, hitStart + stSize - 1) ];
      ScoredSeq* contig = contigLocus->getSeq();
      if (seq != contig and matchTest->seqIsOk(contig) ){ // don't align to self!!!
        // determine the offset for the alignment of the query to this seq;
	// number is the start position of target in seq (i.e. a seq coord)
	long contigStart = hitStart - contigLocus->start();

	// make sure that the hit is on an appropriate region of the target contig
	if ( wordQuery->acceptableOverlap(contigStart, contig) ){
	  long offset = startNum - contigStart;
	  long contigIndex;
	  map<ScoredSeq*,long>::iterator hitIt = hitToIndex.find( contig );
	  if (hitIt == hitToIndex.end()){
	    hitToIndex.insert( pair<ScoredSeq*,long>(contig,currentIndex) );
	    contigIndex = currentIndex;
	    hitArray[contigIndex] = new OffsetTracker(contig);
	    currentIndex++;
	  } else {
	    contigIndex = hitIt->second;
	  }
	  hitArray[contigIndex]->addOffset( offset, numMatches );
	}
      }
    }
    delete wordQuery;
    delete [] matchArray;
    delete [] subtext;
  }

  // now sort by the seed match counts.  first, figure out how many of each number of counts there are
  map<long,long> hitCountsToCount;
  for (long n = 0; n < currentIndex; ++n){
    long maxOffsetCount = hitArray[n]->maxOffsetCount();
    map<long,long>::iterator it = hitCountsToCount.find( maxOffsetCount );
    if (it == hitCountsToCount.end()){ hitCountsToCount.insert( pair<long,long>(maxOffsetCount,1) ); }
    else { it->second++; }
  }

  // now, figure out the index of the sorted list where each seed count #'s sublist will begin
  // decrement since the indexes (num counts) of the map will be retrieved in increasing order
  map<long,long> hitCountsToStart;
  long currentStart = currentIndex;
  for (map<long,long>::iterator it = hitCountsToCount.begin(); it != hitCountsToCount.end(); ++it){
    currentStart -= it->second;
    hitCountsToStart.insert(pair<long,long>(it->first, currentStart));
  }

  // now, create the new sorted list and fill it
  OffsetTracker** sortedHitWithOffsets = new OffsetTracker*[ currentIndex + 1 ];
  for (long n = 0; n < currentIndex; ++n){
    map<long,long>::iterator it = hitCountsToStart.find(hitArray[n]->maxOffsetCount());
    sortedHitWithOffsets[it->second] = hitArray[n];
    it->second++;
  }
  // the last index is a pair of nulls so that I don't have to return a length
  sortedHitWithOffsets[currentIndex] = NULL;
			     
  delete [] queryNumArray;
  delete [] hitArray;
  return sortedHitWithOffsets;
}


void ScoredSeqCollectionBwt::getSubtextWordQueries(ScoredSeq* querySeq, set<WordQuery*>* wordQuerySet, char sense, long minOverlap){

  // this is the whole-sequence case
  long blockStartPos = 0;
  long blockEndPos = querySeq->size();
  getSubtextWordQueriesHelper(querySeq, wordQuerySet, sense, blockStartPos, blockEndPos, false);

  // these are the edge cases
  if (_edgeScalingEnabled){
    long fragmentLen = minOverlap;
    while (fragmentLen < querySeq->size()){
      // the front end
      blockStartPos = 0;
      blockEndPos = fragmentLen;
      getSubtextWordQueriesHelper(querySeq, wordQuerySet, sense, blockStartPos, blockEndPos, true);
      // the back end
      blockStartPos = querySeq->size() - fragmentLen;
      blockEndPos = querySeq->size();
      getSubtextWordQueriesHelper(querySeq, wordQuerySet, sense, blockStartPos, blockEndPos, true);
      // scale down the window
      fragmentLen = fragmentLen * 2;
    }
  }
}



  // num allowed errors:        0  1  2  3  4  5  6  7  8  9  10
int ScoredSeqCollectionBwt::_specialCaseLenDenom[6] =  {1, 2, 3, 3, 4, 5};//, 6, 7, 8, 9, 10};
int ScoredSeqCollectionBwt::_specialCaseStepDenom[6] = {1, 1, 1, 2, 2, 1};//, 1, 1, 1, 1, 1};
int ScoredSeqCollectionBwt::_specialCaseWindowNum[6] = {1, 2, 3, 5, 7, 5};//, 6, 7, 8, 9, 10};

void ScoredSeqCollectionBwt::getSubtextWordQueriesHelper(ScoredSeq* querySeq, set<WordQuery*>* wordQuerySet, char sense,
							 long blockStartPos, long blockEndPos, bool useLimits){
  long seqLen = blockEndPos - blockStartPos;
  bool fiveEdge;
  if (blockStartPos == 0){ fiveEdge = true; }
  else if (blockEndPos == querySeq->size()){ fiveEdge = false; }
  else { throw AssemblyException::ArgError("SSCBwt::getSWQHelper block is at neither the front nor back of the query."); }


  long numAllowedMis = long(float(seqLen) * _maxFractMis);
  // a safety precaution for a rediculous situation
  if (numAllowedMis==seqLen){ numAllowedMis = seqLen - 1; }

  // special cases
  if (numAllowedMis < 6){
    long startDenom = _specialCaseLenDenom[numAllowedMis] * _specialCaseStepDenom[numAllowedMis];
    long subLen = seqLen / _specialCaseLenDenom[numAllowedMis];
    for (long n = 0; n < _specialCaseWindowNum[numAllowedMis]; ++n){
      long start = blockStartPos + (n * seqLen / startDenom );
      WordQuery* wc;
      if (! useLimits){ wc = new WordQuery(querySeq, start, subLen, sense); }
      else { wc = new WordQuery(querySeq, start, subLen, sense, seqLen, fiveEdge); }
      wordQuerySet->insert(wc);
    }
  } else { // general cases
    long numWindows;
    // see notebook7 10/26/10 for derivation
    switch(numAllowedMis){
    case 6: numWindows = 6; break;
    case 7: numWindows = 6; break;
    case 8: numWindows = 7; break;
    case 9: numWindows = 7; break;
    case 10: numWindows = 8; break;
    case 11: numWindows = 8; break;
    case 12: numWindows = 8; break;
    case 13: numWindows = 9; break;
    case 14: numWindows = 9; break;
    case 15: numWindows = 10; break;
    case 16: numWindows = 10; break;
    case 17: numWindows = 11; break;
    case 18: numWindows = 11; break;
    case 19: numWindows = 11; break;
    case 20: numWindows = 12; break;
    case 21: numWindows = 12; break;
    case 22: numWindows = 13; break;
    case 23: numWindows = 13; break;
    case 24: numWindows = 13; break;
    case 25: numWindows = 14; break;
    case 26: numWindows = 14; break;
    case 27: numWindows = 14; break;
    case 28: numWindows = 15; break;
    case 29: numWindows = 15; break;
    case 30: numWindows = 15; break;
    case 31: numWindows = 16; break;
    case 32: numWindows = 16; break;
    case 33: numWindows = 16; break;
    case 34: numWindows = 17; break;
    case 35: numWindows = 17; break;
    case 36: numWindows = 17; break;
    case 37: numWindows = 18; break;
    case 38: numWindows = 18; break;
    case 39: numWindows = 18; break;
    case 40: numWindows = 19; break;
    case 41: numWindows = 19; break;
    case 42: numWindows = 19; break;
    case 43: numWindows = 20; break;
    case 44: numWindows = 20; break;
    case 45: numWindows = 20; break;
    case 46: numWindows = 21; break;
    case 47: numWindows = 21; break;
    case 48: numWindows = 21; break;
    case 49: numWindows = 21; break;
    case 50: numWindows = 22; break;
    case 51: numWindows = 22; break;
    case 52: numWindows = 22; break;
    case 53: numWindows = 23; break;
    case 54: numWindows = 23; break;
    case 55: numWindows = 23; break;
    case 56: numWindows = 24; break;
    case 57: numWindows = 24; break;
    case 58: numWindows = 24; break;
    case 59: numWindows = 24; break;
    case 60: numWindows = 25; break;
    case 61: numWindows = 25; break;
    case 62: numWindows = 25; break;
    case 63: numWindows = 26; break;
    case 64: numWindows = 26; break;
    case 65: numWindows = 26; break;
    case 66: numWindows = 27; break;
    case 67: numWindows = 27; break;
    case 68: numWindows = 27; break;
    case 69: numWindows = 27; break;
    case 70: numWindows = 28; break;
    case 71: numWindows = 28; break;
    case 72: numWindows = 28; break;
    case 73: numWindows = 29; break;
    case 74: numWindows = 29; break;
    case 75: numWindows = 29; break;
    case 76: numWindows = 29; break;
    case 77: numWindows = 30; break;
    case 78: numWindows = 30; break;
    case 79: numWindows = 30; break;
    case 80: numWindows = 31; break;
    case 81: numWindows = 31; break;
    case 82: numWindows = 31; break;
    case 83: numWindows = 31; break;
    case 84: numWindows = 32; break;
    case 85: numWindows = 32; break;
    case 86: numWindows = 32; break;
    case 87: numWindows = 32; break;
    case 88: numWindows = 33; break;
    case 89: numWindows = 33; break;
    case 90: numWindows = 33; break;
    case 91: numWindows = 34; break;
    case 92: numWindows = 34; break;
    case 93: numWindows = 34; break;
    case 94: numWindows = 34; break;
    case 95: numWindows = 35; break;
    case 96: numWindows = 35; break;
    case 97: numWindows = 35; break;
    case 98: numWindows = 35; break;
    case 99: numWindows = 36; break;
    case 100: numWindows = 36; break;
    case 101: numWindows = 36; break;
    case 102: numWindows = 36; break;
    case 103: numWindows = 37; break;
    case 104: numWindows = 37; break;
    case 105: numWindows = 37; break;
    case 106: numWindows = 38; break;
    case 107: numWindows = 38; break;
    case 108: numWindows = 38; break;
    case 109: numWindows = 38; break;
    case 110: numWindows = 39; break;
    case 111: numWindows = 39; break;
    case 112: numWindows = 39; break;
    case 113: numWindows = 39; break;
    case 114: numWindows = 40; break;
    case 115: numWindows = 40; break;
    case 116: numWindows = 40; break;
    case 117: numWindows = 40; break;
    case 118: numWindows = 41; break;
    case 119: numWindows = 41; break;
    case 120: numWindows = 41; break;
    case 121: numWindows = 41; break;
    case 122: numWindows = 42; break;
    case 123: numWindows = 42; break;
    case 124: numWindows = 42; break;
    case 125: numWindows = 42; break;
    case 126: numWindows = 43; break;
    case 127: numWindows = 43; break;
    case 128: numWindows = 43; break;
    case 129: numWindows = 43; break;
    case 130: numWindows = 44; break;
    case 131: numWindows = 44; break;
    case 132: numWindows = 44; break;
    case 133: numWindows = 45; break;
    case 134: numWindows = 45; break;
    case 135: numWindows = 45; break;
    case 136: numWindows = 45; break;
    case 137: numWindows = 46; break;
    case 138: numWindows = 46; break;
    case 139: numWindows = 46; break;
    case 140: numWindows = 46; break;
    default: numWindows = numAllowedMis / 3;
    }
    long subLen = seqLen / numWindows;
    for (long n = 0; n < numWindows; ++n){
      long start = blockStartPos + (n * seqLen / numWindows );
      WordQuery* wc;
      if (! useLimits){ wc = new WordQuery(querySeq, start, subLen, sense); }
      else { wc = new WordQuery(querySeq, start, subLen, sense, seqLen, fiveEdge); }
      wordQuerySet->insert(wc);
    }
  }

  // get rid of word queries with Ns
  set<WordQuery*> wqToRemove;
  for (set<WordQuery*>::iterator it = wordQuerySet->begin(); it != wordQuerySet->end(); ++it){
    WordQuery* wq = *it;
    char* querySeq = wq->_sourceSeq->getSubseq(wq->_start, wq->_size, '+');
    if (string(querySeq).find('N') != string::npos){ wqToRemove.insert(wq); }
    delete [] querySeq;
  }
  for (set<WordQuery*>::iterator it = wqToRemove.begin(); it != wqToRemove.end(); ++it){
    WordQuery* wq = *it;
    wordQuerySet->erase( wq );
    delete wq;
  }
}





void ScoredSeqCollectionBwt::bufferSeqs(){
  //OK();
  for ( set<ScoredSeq*>::iterator it = _allSeqs.begin(); it != _allSeqs.end(); ++it ){ (*it)->buffer(); }
  //OK();
}

void ScoredSeqCollectionBwt::unbufferSeqs(){
  //OK();
  for ( set<ScoredSeq*>::iterator it = _allSeqs.begin(); it != _allSeqs.end(); ++it ){ (*it)->unbuffer(); }
  //OK();
}



long ScoredSeqCollectionBwt::size(){
  //OK();
  return _allSeqs.size();
}


// overlaps is the return value, locus is the query
// ASSUMES that the locus only overlaps one contig
long ScoredSeqCollectionBwt::getOverlapContigIndex(long start, long end){
  set<ScoredSeqLocus*> candOverlaps; // all bin members; must be filtered for true overlaps

  // determine the bin to look in
  long startBin = start / _binSize;
  if (startBin < 0){ startBin = 0; }

  // return the legit match's index
  set<ScoredSeqLocus*>::iterator firstLocusIt = _binToSeqs[startBin]->begin();
  while (firstLocusIt != _binToSeqs[startBin]->end() and (! (*firstLocusIt)->overlaps(start,end))){ ++firstLocusIt; }
  if ( firstLocusIt == _binToSeqs[startBin]->end() ){ throw AssemblyException::LogicError("SSCBwt::gOCI can't find match."); }
  return (*firstLocusIt)->index();
}


ScoredSeqCollectionBwt::ScoredSeqLocus::ScoredSeqLocus(){}
ScoredSeqCollectionBwt::ScoredSeqLocus::ScoredSeqLocus(ScoredSeq* seq, long start, long end, long index) :
  _start(start),
  _end(end),
  _seq(seq),
  _index(index) {
  }
long ScoredSeqCollectionBwt::ScoredSeqLocus::start(){ return _start; }
long ScoredSeqCollectionBwt::ScoredSeqLocus::end(){ return _end; }
long ScoredSeqCollectionBwt::ScoredSeqLocus::index(){ return _index; }
long ScoredSeqCollectionBwt::ScoredSeqLocus::size(){ return _end - _start + 1; }
bool ScoredSeqCollectionBwt::ScoredSeqLocus::overlaps(MatchLocus* otherLocus){
  return (! (_start > otherLocus->end() or otherLocus->start() > _end) );
}
bool ScoredSeqCollectionBwt::ScoredSeqLocus::contains(MatchLocus* otherLocus){
  return (! (_start > otherLocus->start() or otherLocus->end() > _end) );
}
bool ScoredSeqCollectionBwt::ScoredSeqLocus::overlaps(long otherStart, long otherEnd){
  return (! (_start > otherEnd or otherStart > _end) );
}
bool ScoredSeqCollectionBwt::ScoredSeqLocus::contains(long otherStart, long otherEnd){
  return (! (_start > otherStart or otherEnd > _end) );
}
ScoredSeq* ScoredSeqCollectionBwt::ScoredSeqLocus::getSeq(){ return _seq; }




ScoredSeqCollectionBwt::AlignmentMaker::~AlignmentMaker(){}
bool ScoredSeqCollectionBwt::AlignmentMaker::makeAlignmentHelperPlus(ScoredSeq* seqA, long aStart, ScoredSeq* seqB, long bStart,
									   long maxNumMis, long alLength){
  // gaps are irrelevant
  AlignmentScoreMatrix* misCountAsm = new AlignmentScoreMatrix(0,1,1,1);
  long posA = aStart + alLength;
  long posB = bStart + alLength;
  long misCount = 0;
  while ( (maxNumMis >= misCount) and (posA > aStart) ){
    --posA;
    --posB;
    misCount += misCountAsm->match(seqA->nucAtPosPlus(posA), seqB->nucAtPosPlus(posB));
  }
  delete misCountAsm;
  return maxNumMis >= misCount;
}
bool ScoredSeqCollectionBwt::AlignmentMaker::makeAlignmentHelperMinus(ScoredSeq* seqA, long aStart, ScoredSeq* seqB, long bStart,
									    long maxNumMis, long alLength){
  // gaps are irrelevant
  AlignmentScoreMatrix* misCountAsm = new AlignmentScoreMatrix(0,1,1,1);
  long posA = aStart + alLength;
  long posB = bStart + alLength;
  long misCount = 0;
  while ( (maxNumMis >= misCount) and (posA > aStart) ){
    --posA;
    --posB;
    misCount += misCountAsm->match(seqA->nucAtPosPlus(posA), seqB->nucAtPosMinus(posB));
  }
  delete misCountAsm;
  return maxNumMis >= misCount;
}



ScoredSeqCollectionBwt::AlMakerNoEdge::AlMakerNoEdge(){}
ScoredSeqCollectionBwt::AlMakerNoEdge::AlMakerNoEdge(AlignmentScoreMatrix* scoreMatrix, long minScore, float maxFractMis) :
  _scoreMatrix(scoreMatrix),
  _minScore(minScore),
  _maxFractMis(maxFractMis)
{
  if (_scoreMatrix->_match <= _scoreMatrix->_mismatch){
    throw AssemblyException::ArgError("SSCBwt::AlMakerNoEdge needs a lower mismatch score than match score (i.e. mismatch should be negative)");
  }
}
ScoredSeqCollectionBwt::AlMakerNoEdge::~AlMakerNoEdge(){}
Alignment* ScoredSeqCollectionBwt::AlMakerNoEdge::makeAlignment(ScoredSeq* seqA, ScoredSeq* seqB, char sense, long offset){
  // get the starts and alignment length
  long aStart;
  long bStart;
  if (offset < 0){
    aStart = 0;
    bStart = 0 - offset;
  } else {
    aStart = offset;
    bStart = 0;
  }
  long aEnd;
  if (offset + seqB->size() >= seqA->size()){ aEnd = seqA->size(); }
  else { aEnd = offset + seqB->size(); }
  long alLength = aEnd - aStart;

  // i can maybe take out these tests
  if (aStart != 0 and bStart != 0){
    throw AssemblyException::ArgError("in SSCBwt::AlMaker, neither start is zero");
  }
  if ( aStart + alLength != seqA->size() and bStart + alLength != seqB->size() ){
    throw AssemblyException::ArgError("in SSCBwt::AlMaker, neither end is the end of the sequence");
  }

  // the min score must be beaten, and the fractId must also be satisfied
  long maxNumMisByScore = (_minScore - _scoreMatrix->_match * alLength) / (_scoreMatrix->_mismatch - _scoreMatrix->_match );
  long maxNumMisByFid = long(float(alLength) * _maxFractMis);
  long maxNumMis;
  if (maxNumMisByScore < maxNumMisByFid){ maxNumMis = maxNumMisByScore; }
  else { maxNumMis = maxNumMisByFid; }

  bool alIsGood;
  switch (sense){
  case '+': alIsGood = makeAlignmentHelperPlus(seqA, aStart, seqB, bStart, maxNumMis, alLength); break;
  case '-': alIsGood = makeAlignmentHelperMinus(seqA, aStart, seqB, bStart, maxNumMis, alLength); break;
  default: throw AssemblyException::ArgError("SSCBwt:AlMakerNoEdge:makeAl sense char is bad");
  }
  if (alIsGood){ return new AlignmentUngapped(seqA, seqB, sense, offset); }
  else { return new AlignmentNull(seqA, seqB, sense); }
}
long ScoredSeqCollectionBwt::AlMakerNoEdge::getScore(Alignment* al){
  return al->score(_scoreMatrix, false);
}
long ScoredSeqCollectionBwt::AlMakerNoEdge::getMinScore(){ return _minScore; }
void ScoredSeqCollectionBwt::AlMakerNoEdge::setMinScore(long minScore){ _minScore = minScore; }
AlignmentScoreMatrix* ScoredSeqCollectionBwt::AlMakerNoEdge::getScoreMatrix(){ return _scoreMatrix; }



ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::AlMakerPenalizeEdgeA(){}
ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::AlMakerPenalizeEdgeA(AlignmentScoreMatrix* scoreMatrix, long minScore, float maxFractMis) :
  _scoreMatrix(scoreMatrix),
  _minScore(minScore),
  _maxFractMis(maxFractMis)
{
  if (_scoreMatrix->_match <= _scoreMatrix->_mismatch){
    throw AssemblyException::ArgError("SSCBwt::AlMakerNoEdge needs a lower mismatch score than match score (i.e. mismatch should be negative)");
  }
}
ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::~AlMakerPenalizeEdgeA(){}

Alignment* ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::makeAlignment(ScoredSeq* seqA, ScoredSeq* seqB, char sense, long offset){
  // get the starts and alignment length
  long aStart;
  long bStart;
  if (offset < 0){
    aStart = 0;
    bStart = 0 - offset;
  } else {
    aStart = offset;
    bStart = 0;
  }
  long aEnd;
  if (offset + seqB->size() >= seqA->size()){ aEnd = seqA->size(); }
  else { aEnd = offset + seqB->size(); }
  long alLength = aEnd - aStart;

  // i can maybe take out these tests
  if (aStart != 0 and bStart != 0){
    throw AssemblyException::ArgError("in SSCBwt::AlMaker, neither start is zero");
  }
  if ( aStart + alLength != seqA->size() and bStart + alLength != seqB->size() ){
    throw AssemblyException::ArgError("in SSCBwt::AlMaker, neither end is the end of the sequence");
  }

  // the min score must be beaten, and the fractId must also be satisfied
  long maxNumMisByScore = (_minScore - _scoreMatrix->_match * seqA->size()) / (_scoreMatrix->_mismatch - _scoreMatrix->_match );
  long maxNumMisByFid = long(float(seqA->size()) * _maxFractMis);
  long maxNumMis;
  if (maxNumMisByScore < maxNumMisByFid){ maxNumMis = maxNumMisByScore; }
  else { maxNumMis = maxNumMisByFid; }
  // adjust for the edges
  maxNumMis -= seqA->size() - alLength;

  bool alIsGood;
  switch (sense){
  case '+': alIsGood = makeAlignmentHelperPlus(seqA, aStart, seqB, bStart, maxNumMis, alLength); break;
  case '-': alIsGood = makeAlignmentHelperMinus(seqA, aStart, seqB, bStart, maxNumMis, alLength); break;
  default: throw AssemblyException::ArgError("SSCBwt:AlMakerNoEdge:makeAl sense char is bad");
  }
  if (alIsGood){ return new AlignmentUngapped(seqA, seqB, sense, offset); }
  else { return new AlignmentNull(seqA, seqB, sense); }
}
long ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::getScore(Alignment* al){
  return al->scoreOverhangA(_scoreMatrix);
}
long ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::getMinScore(){ return _minScore; }
void ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::setMinScore(long minScore){ _minScore = minScore; }
AlignmentScoreMatrix* ScoredSeqCollectionBwt::AlMakerPenalizeEdgeA::getScoreMatrix(){ return _scoreMatrix; }




ScoredSeqCollectionBwt::OffsetTracker::OffsetTracker(ScoredSeq* target) : _targetContig(target){
  _maxOffsetCount = 0;
  // this value is meaningless unless the above value is >0
  _maxCountOffset = 0;
}
ScoredSeqCollectionBwt::OffsetTracker::~OffsetTracker(){}
void ScoredSeqCollectionBwt::OffsetTracker::addOffset(long offset, long normFactor){
  float localCount;
  if (normFactor < 1){ throw AssemblyException::ArgError("SSCBwt::OffsetTracker::addOffset, normFactor cannot be less than 1"); }
  float localAddition = float(1) / float(normFactor);
  // now find or create the offset entry
  map<long,float>::iterator offsetEntry = _offsetToCount.find(offset);
  if (offsetEntry == _offsetToCount.end()){
    localCount = localAddition;
    _offsetToCount.insert( pair<long,float>(offset,localCount) );
  } else {
    offsetEntry->second += localAddition;
    localCount = offsetEntry->second;
  }
  if (localCount > _maxOffsetCount){
    _maxOffsetCount = localCount;
    _maxCountOffset = offset;
  }
}
long ScoredSeqCollectionBwt::OffsetTracker::maxOffsetCount(){ return _maxOffsetCount; }
long ScoredSeqCollectionBwt::OffsetTracker::maxCountOffset(){ return _maxCountOffset; }
long* ScoredSeqCollectionBwt::OffsetTracker::getOffsets(){
  // figure out how many offsets there are with each count
  map<float,long> countToNum;
  for (map<long,float>::iterator it = _offsetToCount.begin(); it != _offsetToCount.end(); ++it){
    map<float,long>::iterator foundCtm = countToNum.find( it->second );
    if (foundCtm == countToNum.end()){ countToNum.insert(pair<float,long>(it->second,1)); }
    else { foundCtm->second += 1; }
  }
  // figure out the index in the sorted array at which each offset count will begin
  map<float,long> countToIndex;
  long index = _offsetToCount.size();
  for (map<float,long>::iterator it = countToNum.begin(); it != countToNum.end(); ++it){
    index -= it->second;
    countToIndex.insert(pair<float,long>(it->first,index));
  }
  // fill in the array so it is sorted
  long* offsetArray = new long[_offsetToCount.size() + 1];
  for (map<long,float>::iterator it = _offsetToCount.begin(); it != _offsetToCount.end(); ++it){
    map<float,long>::iterator foundCtm = countToIndex.find( it->second );
    offsetArray[ foundCtm->second ] = it->first;
    foundCtm->second += 1;
  }
  return offsetArray;
}

long ScoredSeqCollectionBwt::OffsetTracker::numOffsets(){ return long(_offsetToCount.size()); }

ScoredSeq* ScoredSeqCollectionBwt::OffsetTracker::targetContig(){ return _targetContig; }


ScoredSeqCollectionBwt::WordQuery::WordQuery(ScoredSeq* sourceSeq, long start, long size, char sense) :
  _sourceSeq(sourceSeq), _start(start), _size(size), _sense(sense), 
  _usingEdgeLimit(false), _edgeLimit(-1), _fiveEdge(true) {}
ScoredSeqCollectionBwt::WordQuery::WordQuery(ScoredSeq* sourceSeq, long start, long size, char sense, long edgeLimit, bool fiveEdge) :
  _sourceSeq(sourceSeq), _start(start), _size(size), _sense(sense), 
  _usingEdgeLimit(true), _edgeLimit(edgeLimit), _fiveEdge(fiveEdge) {}
bool ScoredSeqCollectionBwt::WordQuery::acceptableOverlap(long targetStart, ScoredSeq* target){
  if (! _usingEdgeLimit){ return true; }
  else {
    if ( _fiveEdge ){ // target's 3p edge
      long distance = target->size() - targetStart;
      return distance <= _edgeLimit * 2;
    } else { // target's 5p edge
      long distance = targetStart + _size;
      return distance <= _edgeLimit * 2;
    }
  }
}



#endif


