/*
PRICE was written by Graham Ruby.
This file is part of PRICE.

PRICE is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

PRICE is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with PRICE.  If not, see <http://www.gnu.org/licenses/>.
*/

/* A collection of ScoredSeq objects indexed for quick retrieval

This version is optimized for high speed when high fractId is required.
The set of ScoredSeqs that it contains cannot be modified.

The contents of this collection are typed.
 */


#ifndef SCOREDSEQCOLLECTIONBWT_H
#define SCOREDSEQCOLLECTIONBWT_H

#include <omp.h>
#include <set>
#include <map>
#include "ScoredSeq.h"
#include "Alignment.h"
#include "DynamicProgrammingAligner.h"
#include "AlignmentScoreMatrix.h"
#include "burrowsWheelerUtilities.h"
#include "MatchSeqTest.h"
#include "MatchOffsetTest.h"
#include "MatchAlignTest.h"
using namespace::std;


class ScoredSeqCollectionBwt{

 public:
  ScoredSeqCollectionBwt();
  // these two will be used if ungapped alignments are to be returned
  ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, float fractId);
  ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, float fractId, long minOverlap);
  ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, float fractId, long minOverlap, long maxOverlap);
  // these two will be used if gapped alignments are to be returned
  ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, DynamicProgrammingAligner * asi);
  ScoredSeqCollectionBwt(set<ScoredSeq*>* inputSeqs, DynamicProgrammingAligner * asi, long globalMaxOverlap);
  ~ScoredSeqCollectionBwt();

  enum GapMode { gapped=0, ungapped=1 };
  enum MinOvlStringency { hardMinOvl=0, softMinOvl=1 };

  // DEFAULT: edge scaling is enabled
  // that means that matches to smaller words from the edges of query contigs
  // will be sought in order to find matches with small overlaps.
  void enableEdgeScaling();
  void disableEdgeScaling();

  // returns a deep copy of the collection that uses only the
  // shallow versions of the contained objects so that there are
  // no pointers that could be used in some other part of the
  // software (excepting singletons).
  ScoredSeqCollectionBwt * copy();

  // methods for dealing with contents
  bool contains(ScoredSeq* s);
  void getSeqs( set<ScoredSeq*>* );

  // learn about the collection
  float getFractId();

  // In the methods below, a pre-created set of alignments is passed in as a reference.
  // It will be modified by having the new matches added to it.

  // these methods return just the best matches
  // (according to score; all equal matches are returned)
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq);  // both senses are returned
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense);  // sense can be '+' or '-'
  // and these have restricted targets
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* seqTest);  // both senses are returned
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* seqTest);  // sense can be '+' or '-'

  // now those with restricted alignments
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchAlignTest* alignTest);
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchAlignTest* alignTest);
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* seqTest, MatchAlignTest* alignTest);
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* seqTest, MatchAlignTest* alignTest);

  // these allow a new minOverlap to be set
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq,
		      long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // both senses are returned
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense,
		      long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // sense can be '+' or '-'
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* seqTest,
		      long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // both senses are returned
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* seqTest,
		      long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // sense can be '+' or '-'

  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchAlignTest* alignTest,
		      long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // both senses are returned
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchAlignTest* alignTest,
		      long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // sense can be '+' or '-'
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* seqTest, MatchAlignTest* alignTest,
		      long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // both senses are returned
  void getBestMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* seqTest, MatchAlignTest* alignTest,
		      long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // sense can be '+' or '-'


  // these methods return just the best matches
  // (according to score; all equal matches are returned)
  bool hasMatch(ScoredSeq* seq);  // both senses are returned
  bool hasMatch(ScoredSeq* seq, char sense);  // sense can be '+' or '-'
  // and these have restricted targets
  bool hasMatch(ScoredSeq* seq, MatchSeqTest* seqTest);  // both senses are returned
  bool hasMatch(ScoredSeq* seq, char sense, MatchSeqTest* seqTest);  // sense can be '+' or '-'

  // now those with restricted alignments
  bool hasMatch(ScoredSeq* seq, MatchAlignTest* alignTest);
  bool hasMatch(ScoredSeq* seq, char sense, MatchAlignTest* alignTest);
  bool hasMatch(ScoredSeq* seq, MatchSeqTest* seqTest, MatchAlignTest* alignTest);
  bool hasMatch(ScoredSeq* seq, char sense, MatchSeqTest* seqTest, MatchAlignTest* alignTest);

  // these allow a new minOverlap to be set
  bool hasMatch(ScoredSeq* seq, long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // both senses are returned
  bool hasMatch(ScoredSeq* seq, char sense, long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // sense can be '+' or '-'
  bool hasMatch(ScoredSeq* seq, MatchSeqTest* seqTest, long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // both senses are returned
  bool hasMatch(ScoredSeq* seq, char sense, MatchSeqTest* seqTest, long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // sense can be '+' or '-'

  bool hasMatch(ScoredSeq* seq, MatchAlignTest* alignTest, long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);
  bool hasMatch(ScoredSeq* seq, char sense, MatchAlignTest* alignTest, long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);
  bool hasMatch(ScoredSeq* seq, MatchSeqTest* seqTest, MatchAlignTest* alignTest, long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);
  bool hasMatch(ScoredSeq* seq, char sense, MatchSeqTest* seqTest, MatchAlignTest* alignTest, long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);


  // and i have equivalents that return all matches
  void getMatches(set<Alignment*>* matches, ScoredSeq* seq);  // both senses are returned
  void getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense);  // sense can be '+' or '-'
  void getMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* matchTest);  // both senses are returned
  void getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* matchTest);  // sense can be '+' or '-'
  void getMatches(set<Alignment*>* matches, ScoredSeq* seq,
		  long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // both senses are returned
  void getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense,
		  long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // sense can be '+' or '-'
  void getMatches(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* matchTest,
		  long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // both senses are returned
  void getMatches(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* matchTest,
		  long minOverlap, MinOvlStringency ovlStringency = hardMinOvl);  // sense can be '+' or '-'


  // these two methods deal with the buffering for all the member seqs.  they 
  // don't change the abstract state; they are provided as tools for saving memory
  // without destroying the collection; their converse operations are private and 
  // will be called based on demand (i.e. when matches are requested).
  void unbufferSeqs();
  void unindexSeqs();

  long size(); // number of member ScoredSeqs
  void OK();

  void bufferSeqs();

  // this is only public so that I can test it
  //Alignment* reverseAlignmentSeqA(Alignment* oldAl, ScoredSeq* newSeqA);



 private:
  void constructorHelper(set<ScoredSeq*>* inputSeqs);
  bool _penalizeEdgeGaps;

  // this should be true during assembly and false during read mapping
  bool _edgeScalingEnabled;

  // private versions of these methods; the above methods will forward to them
  void getBestMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq,
			     MatchSeqTest* seqTest, MatchAlignTest* alignTest,
			     long minOverlap);
  void getBestMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq, char sense,
			     MatchSeqTest* seqTest, MatchAlignTest* alignTest,
			     long minOverlap);

  // private versions of these methods; the above methods will forward to them
  bool hasMatchPrivate(ScoredSeq* seq,
		       MatchSeqTest* seqTest, MatchAlignTest* alignTest,
		       long minOverlap);
  bool hasMatchPrivate(ScoredSeq* seq, char sense,
		       MatchSeqTest* seqTest, MatchAlignTest* alignTest,
		       long minOverlap);


  void getMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq, MatchSeqTest* matchTest, long minOverlap);
  void getMatchesPrivate(set<Alignment*>* matches, ScoredSeq* seq, char sense, MatchSeqTest* matchTest, long minOverlap);

  // these methods create an internal representation with a bigger memory
  // footprint and so are called based on demand (i.e. when matches are 
  // requested).
  void indexSeqs();
  void indexSeq(ScoredSeq * seq); // some functional abstraction

  float _fractId;
  float _maxFractMis;
  long _maxOverlap;
  long _minOverlap;
  bool _usingMaxOverlap;

  GapMode _gapMode;


  // these are for the gapped alignment mode
  DynamicProgrammingAligner * _asi;


  // BWT stuff
  long* _sortedSuffixes; // also an array of text length
  long* _bwTransform; // also an array of text length
  long* _occCounts; // also an array of text length
  long* _tableC; // an array of alphabet length
  long _textSize;

  // I NEED A CLASS OF SCOREDSEQ WRAPPERS THAT DOES THE OVERLAP FUNCTION AND KEEPS TRACK
  // OF COORD POSITION IN THE META-TEXT USED FOR BWT MAPPING.

  // used to query for the hit ScoredSeq
  class MatchLocus {
  public:
    virtual long start() = 0;
    virtual long end() = 0; // end = position of the last letter, NOT the +1
    virtual long size() = 0;
    virtual bool overlaps(MatchLocus* otherLocus) = 0;
    virtual bool contains(MatchLocus* otherLocus) = 0;
    virtual bool overlaps(long otherStart, long otherEnd) = 0;
    virtual bool contains(long otherStart, long otherEnd) = 0;
  };

  // used to store the ScoredSeqs so they can be queried
  class ScoredSeqLocus {
  public:
    ScoredSeqLocus();
    ScoredSeqLocus(ScoredSeq* seq, long start, long end, long index);
    long start();
    long end(); // end = position of the last letter, NOT the +1
    long size();
    long index(); // CLASS-SPECIFIC
    bool overlaps(MatchLocus* otherLocus);
    bool contains(MatchLocus* otherLocus);
    bool overlaps(long otherStart, long otherEnd);
    bool contains(long otherStart, long otherEnd);
    ScoredSeq* getSeq();
  private:
    ScoredSeq* _seq;
    long _start;
    long _end;
    long _index;
  };


  // for keeping track of offsets and passing them around
  class OffsetTracker {
  public:
    OffsetTracker(ScoredSeq* target);
    ~OffsetTracker();
    void addOffset(long offset, long normFactor=1);
    long maxOffsetCount();
    long maxCountOffset();
    long* getOffsets();
    long numOffsets();
    ScoredSeq* targetContig();
  private:
    map<long,float> _offsetToCount;
    long _maxCountOffset;
    float _maxOffsetCount;
    ScoredSeq* _targetContig;
  };


  // pure abstract class that will replace the one above; I will change
  // the name when I am done with it
  class AlignmentMaker {
  public:
    virtual ~AlignmentMaker();
    virtual Alignment* makeAlignment(ScoredSeq* seqA, ScoredSeq* seqB, char sense, long offset) = 0;
    virtual long getScore(Alignment* al) = 0;
    virtual long getMinScore() = 0;
    virtual void setMinScore(long minScore) = 0;
    virtual AlignmentScoreMatrix* getScoreMatrix() = 0;
    static bool makeAlignmentHelperPlus(ScoredSeq* seqA, long aStart, ScoredSeq* seqB, long bStart, long maxNumMis, long alLength);
    static bool makeAlignmentHelperMinus(ScoredSeq* seqA, long aStart, ScoredSeq* seqB, long bStart, long maxNumMis, long alLength);
  };
  // implementing classes
  class AlMakerNoEdge : public AlignmentMaker {
  public:
    AlMakerNoEdge();
    AlMakerNoEdge(AlignmentScoreMatrix* scoreMatrix, long minScore, float maxFractMis);
    ~AlMakerNoEdge();
    Alignment* makeAlignment(ScoredSeq* seqA, ScoredSeq* seqB, char sense, long offset);
    long getScore(Alignment* al);
    long getMinScore();
    void setMinScore(long minScore);
    AlignmentScoreMatrix* getScoreMatrix();
  private:
    AlignmentScoreMatrix* _scoreMatrix;
    long _minScore;
    float _maxFractMis;
  };
  class AlMakerPenalizeEdgeA : public AlignmentMaker {
  public:
    AlMakerPenalizeEdgeA();
    AlMakerPenalizeEdgeA(AlignmentScoreMatrix* scoreMatrix, long minScore, float maxFractMis);
    ~AlMakerPenalizeEdgeA();
    Alignment* makeAlignment(ScoredSeq* seqA, ScoredSeq* seqB, char sense, long offset);
    long getScore(Alignment* al);
    long getMinScore();
    void setMinScore(long minScore);
    AlignmentScoreMatrix* getScoreMatrix();
  private:
    AlignmentScoreMatrix* _scoreMatrix;
    long _minScore;
    float _maxFractMis;
  };

  // helps enforce the limits of offsets that may be generated from
  // a given query's word match
  class WordQuery{
  public:
    WordQuery(ScoredSeq* sourceSeq, long start, long size, char sense);
    WordQuery(ScoredSeq* sourceSeq, long start, long size, char sense, long edgeLimit, bool fiveEdge);
    // define the souce seq and position on it of the query word
    ScoredSeq* _sourceSeq;
    long _start;
    long _size;
    char _sense;
    // define the distance to the edge of a contig within
    // which the word match must be found (or its relevance)
    bool _usingEdgeLimit;
    long _edgeLimit;
    bool _fiveEdge;
    bool acceptableOverlap(long targetStart, ScoredSeq* target);
  };

  // I will use this to find the ScoredSeqs to which queries were mapped quickly
  // indexes are 0 -> total text length / _binSize (continuous)
  // this will be an array of ScoredSeqLocus* sets.
  set<ScoredSeqLocus*>** _binToSeqs;

  // an array of loci
  ScoredSeqLocus** _orderedContigLoci;
  long _numContigLoci;

  long _binSize;
  long _maxBin; // for safety checks
  long getOverlapContigIndex(long start, long end);
  AlignmentScoreMatrix* _matchCountAsm;
  AlignmentScoreMatrix* _misCountAsm; // gaps count as mismatches here (terminal gaps will exist)
  AlignmentScoreMatrix* _scoreAsm;

  // helpers for functions above that use private datatypes
  void getBestMatchesPrivateUngapped(OffsetTracker** contigToOffsets,
				     set<Alignment*>* matches, ScoredSeq* seq, char sense,
				     MatchSeqTest* seqTest, MatchAlignTest* alignTest, long minOverlap, bool justFirstMatch=false);
  void getBestMatchesPrivateGapped(OffsetTracker** contigToOffsets,
				   set<Alignment*>* matches, ScoredSeq* seq, char sense,
				   MatchSeqTest* seqTest, MatchAlignTest* alignTest,
				   long minOverlap, long bestScoreSoFar, bool justFirstMatch=false);

  // used to define the set of subtext strings for which perfect matches will be sought
  static int _specialCaseLenDenom[6];
  static int _specialCaseStepDenom[6];
  static int _specialCaseWindowNum[6];
  void getSubtextWordQueries(ScoredSeq* querySeq, set<WordQuery*>* wordQuerySet, char sense, long minOverlap);
  void getSubtextWordQueriesHelper(ScoredSeq* querySeq, set<WordQuery*>* wordQuerySet, char sense,
				   long blockStartPos, long blockEndPos, bool useLimits);

  // a helper for getMatches/getBestMatches
  OffsetTracker** getOffsets(ScoredSeq* seq, char sense, long minOverlap, MatchSeqTest* matchTest);


  // these keep track of the total content, binned by whether or not the
  // work that needs to be done to get the collection ready to use has
  // already been performed.
  set<ScoredSeq*> _allSeqs;

};

#endif
